/*
 * SPDX-License-Identifier: AGPL-3.0-only
 * Copyright 2005 - 2016 Zarafa and its licensors
 */
#pragma once
#include <memory>
#include <mutex>
#include "m4l.mapidefs.h"
#include "m4l.mapisvc.h"
#include <mapix.h>
#include <mapispi.h>
#include <string>
#include <list>
#include <kopano/ECConfig.h>
#include <kopano/ECUnknown.h>
#include <kopano/memory.hpp>
#include <kopano/zcdefs.h>

class M4LMsgServiceAdmin;
namespace KC {
class SessionRestorer;
}

struct providerEntry {
	MAPIUID uid;
	std::string servicename; // this provider belongs to service 'servicename'
	KC::object_ptr<M4LProfSect> profilesection;
};

struct serviceEntry {
    MAPIUID muid;
	std::string servicename, displayname;
	KC::object_ptr<M4LProviderAdmin> provideradmin;
	bool bInitialize;
	std::shared_ptr<SVCService> service;
};

struct profEntry {
	std::string profname, password;
	KC::object_ptr<M4LMsgServiceAdmin> serviceadmin;
};

class M4LProfAdmin final : public KC::ECUnknown, public IProfAdmin {
private:
    // variables
	std::list<std::unique_ptr<profEntry> > profiles;
	std::recursive_mutex m_mutexProfiles;

    // functions
    decltype(profiles)::iterator findProfile(const TCHAR *name);

public:
	virtual HRESULT GetLastError(HRESULT, unsigned int flags, MAPIERROR **) override;
	virtual HRESULT GetProfileTable(unsigned int flags, IMAPITable **) override;
	virtual HRESULT CreateProfile(const TCHAR *name, const TCHAR *password, ULONG_PTR ui_param, unsigned int flags) override;
	virtual HRESULT DeleteProfile(const TCHAR *name, unsigned int flags) override;
	virtual HRESULT AdminServices(const TCHAR *name, const TCHAR *password, ULONG_PTR ui_param, unsigned int flags, IMsgServiceAdmin **) override;
	virtual HRESULT QueryInterface(const IID &, void **) override;

	friend class KC::SessionRestorer;
};

class M4LMsgServiceAdmin KC_FINAL_OPG :
    public KC::ECUnknown, public IMsgServiceAdmin2 {
private:
	std::list<std::unique_ptr<providerEntry> > providers;
	std::list<std::unique_ptr<serviceEntry> > services;
	KC::object_ptr<M4LProfSect> profilesection; // Global Profile Section
	/* guards content in service (and provider) list */
	std::recursive_mutex m_mutexserviceadmin;

    // functions
	serviceEntry *findServiceAdmin(const TCHAR *name);
	serviceEntry *findServiceAdmin(const MAPIUID *id);
	providerEntry *findProvider(const MAPIUID *id);

public:
	M4LMsgServiceAdmin(KC::object_ptr<M4LProfSect> &&);
	~M4LMsgServiceAdmin();
	virtual HRESULT GetLastError(HRESULT, unsigned int flags, MAPIERROR **) override;
	virtual HRESULT GetMsgServiceTable(unsigned int flags, IMAPITable **) override;
	virtual HRESULT CreateMsgService(const TCHAR *service, const TCHAR *display_name, ULONG_PTR ui_param, unsigned int flags) override;
	virtual HRESULT CreateMsgServiceEx(const char *service, const char *display_name, ULONG_PTR ui_param, unsigned int flags, MAPIUID *out) override;
	virtual HRESULT DeleteMsgService(const MAPIUID *uid) override;
	virtual HRESULT ConfigureMsgService(const MAPIUID *uid, ULONG_PTR ui_param, unsigned int flags, unsigned int nvals, const SPropValue *props) override;
	virtual HRESULT OpenProfileSection(const MAPIUID *uid, const IID *intf, unsigned int flags, IProfSect **) override;
	virtual HRESULT AdminProviders(const MAPIUID *uid, unsigned int flags, IProviderAdmin **) override;
	virtual HRESULT GetProviderTable(unsigned int flags, IMAPITable **) override;
	virtual HRESULT QueryInterface(const IID &, void **) override;

	friend class M4LProviderAdmin;
	friend class M4LMAPISession;
	friend class KC::SessionRestorer;
};

inline bool operator<(const GUID &a, const GUID &b) noexcept
{
    return memcmp(&a, &b, sizeof(GUID)) < 0;
}

class M4LMAPISession KC_FINAL_OPG : public KC::ECUnknown, public IMAPISession {
private:
	// variables
	std::string profileName;
	KC::object_ptr<M4LMsgServiceAdmin> serviceAdmin;

public:
	M4LMAPISession(const TCHAR *profname, M4LMsgServiceAdmin *);
	virtual HRESULT GetLastError(HRESULT, unsigned int flags, MAPIERROR **) override;
	virtual HRESULT GetMsgStoresTable(unsigned int flags, IMAPITable **) override;
	virtual HRESULT OpenMsgStore(ULONG_PTR ui_param, ULONG eid_size, const ENTRYID *, const IID *intf, ULONG flags, IMsgStore **) override;
	virtual HRESULT OpenAddressBook(ULONG_PTR ui_param, const IID *intf, unsigned int flags, IAddrBook **) override;
	virtual HRESULT OpenProfileSection(const MAPIUID *uid, const IID *intf, unsigned int flags, IProfSect **) override;
	virtual HRESULT OpenEntry(unsigned int eid_size, const ENTRYID *eid, const IID *intf, unsigned int flags, unsigned int *obj_type, IUnknown **) override;
	virtual HRESULT CompareEntryIDs(unsigned int asize, const ENTRYID *a, unsigned int bsize, const ENTRYID *b, unsigned int cmp_flags, unsigned int *result) override;
	virtual HRESULT Advise(ULONG eid_size, const ENTRYID *eid, ULONG evt_mask, IMAPIAdviseSink *, ULONG *conn_id) override;
	virtual HRESULT Unadvise(unsigned int conn) override;
	virtual HRESULT QueryIdentity(unsigned int *eid_size, ENTRYID **) override;
	virtual HRESULT Logoff(ULONG_PTR ui_param, unsigned int flags, unsigned int unused) override;
	virtual HRESULT AdminServices(unsigned int flags, IMsgServiceAdmin **) override;
	virtual HRESULT QueryInterface(const IID &, void **) override;

private:
	/* @todo need a status row per provider */
	ULONG m_cValuesStatus = 0;
	KC::memory_ptr<SPropValue> m_lpPropsStatus;
	std::mutex m_mutexStatusRow;

public:
	HRESULT setStatusRow(ULONG nvals, const SPropValue *);
};

class M4LAddrBook KC_FINAL_OPG : public M4LMAPIProp, public IAddrBook {
public:
	M4LAddrBook(M4LMsgServiceAdmin *new_sa, IMAPISupport *new_sup);
	virtual ~M4LAddrBook();
	virtual HRESULT OpenEntry(unsigned int eid_size, const ENTRYID *eid, const IID *intf, unsigned int flags, unsigned int *obj_type, IUnknown **) override;
	virtual HRESULT CompareEntryIDs(unsigned int asize, const ENTRYID *a, unsigned int bsize, const ENTRYID *b, unsigned int cmp_flags, unsigned int *result) override;
	virtual HRESULT CreateOneOff(const TCHAR *name, const TCHAR *addrtype, const TCHAR *addr, ULONG flags, ULONG *eid_size, ENTRYID **) override;
	virtual HRESULT ResolveName(ULONG_PTR ui_param, ULONG flags, const TCHAR *new_title, ADRLIST *adrlist) override;
	virtual HRESULT GetDefaultDir(unsigned int *eid_size, ENTRYID **) override;
	virtual HRESULT GetSearchPath(unsigned int flags, SRowSet **) override;
	virtual HRESULT SetSearchPath(ULONG flags, const SRowSet *) override;
	virtual HRESULT PrepareRecips(unsigned int flags, const SPropTagArray *, ADRLIST *recips) override;
	virtual HRESULT QueryInterface(const IID &, void **) override;

private:
	// variables
	KC::object_ptr<IMAPISupport> m_lpMAPISup;
	std::list<abEntry> m_lABProviders;
	SRowSet *m_lpSavedSearchPath = nullptr;
	HRESULT getDefaultSearchPath(ULONG ulFlags, LPSRowSet* lppSearchPath);

public:
	HRESULT addProvider(const std::string &profile, const std::string &displayname, MAPIUID *, IABProvider *);
};

extern KC::ECConfig *m4l_lpConfig;
