 /*
 * SPDX-License-Identifier: AGPL-3.0-only
 * Copyright 2005 - 2016 Zarafa and its licensors
 */
#include <memory>
#include <mutex>
#include <string>
#include <climits>
#include <kopano/platform.h>
#include <kopano/memory.hpp>
#include "m4l.mapispi.h"
#include "m4l.mapix.h"
#include <mapi.h>
#include <kopano/CommonUtil.h>
#include <kopano/Util.h>
#include <kopano/ECGuid.h>
#include <kopano/ECUnknown.h>
#include <kopano/mapiguidext.h>
#include <algorithm>
#include <set>
#include <utility>

using namespace KC;

M4LMAPIGetSession::M4LMAPIGetSession(IMAPISession *new_session) :
	session(new_session)
{
	assert(new_session != NULL);
}

HRESULT M4LMAPIGetSession::GetMAPISession(LPUNKNOWN *lppSession)
{
	return session->QueryInterface(IID_IMAPISession, reinterpret_cast<void **>(lppSession));
}

HRESULT M4LMAPIGetSession::QueryInterface(const IID &refiid, void **lppInterface)
{
	REGISTER_INTERFACE2(IMAPIGetSession, this);
	REGISTER_INTERFACE2(IUnknown, this);
	return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

M4LMAPISupport::M4LMAPISupport(IMAPISession *new_session, MAPIUID *lpUid,
    std::shared_ptr<SVCService> &&lpService) :
	session(new_session), service(std::move(lpService))
{
	if(lpUid) {
		lpsProviderUID.reset(new MAPIUID);
		memcpy(lpsProviderUID.get(), lpUid, sizeof(MAPIUID));
	}
}

M4LMAPISupport::~M4LMAPISupport() {
	for (const auto &i : m_advises)
		MAPIFreeBuffer(i.second.lpKey);
}

HRESULT M4LMAPISupport::GetMemAllocRoutines(ALLOCATEBUFFER **lpAllocateBuffer,
    ALLOCATEMORE **lpAllocateMore, FREEBUFFER **lpFreeBuffer)
{
	*lpAllocateBuffer = MAPIAllocateBuffer;
	*lpAllocateMore   = MAPIAllocateMore;
	*lpFreeBuffer     = MAPIFreeBuffer;
	return hrSuccess;
}

HRESULT M4LMAPISupport::Subscribe(const NOTIFKEY *lpKey, ULONG ulEventMask,
    ULONG ulFlags, IMAPIAdviseSink *lpAdviseSink, ULONG *lpulConnection)
{
	ulock_normal l_adv(m_advises_mutex, std::defer_lock_t());

	/* Copy key (this should prevent deletion of the key while it is still in the list */
	memory_ptr<NOTIFKEY> lpNewKey;
	auto hr = MAPIAllocateBuffer(CbNewNOTIFKEY(sizeof(GUID)), &~lpNewKey);
	if (hr != hrSuccess)
		return hr;

	memcpy(lpNewKey, lpKey, sizeof(*lpKey));
	l_adv.lock();
	++m_connections;
	m_advises.emplace(m_connections, M4LSUPPORTADVISE(std::move(lpNewKey), ulEventMask, ulFlags, lpAdviseSink));
	*lpulConnection = m_connections;
	return hrSuccess;
}

HRESULT M4LMAPISupport::Unsubscribe(ULONG ulConnection) {
	scoped_lock l_adv(m_advises_mutex);
	auto i = m_advises.find(ulConnection);
	if (i == m_advises.cend())
		return MAPI_E_NOT_FOUND;
	MAPIFreeBuffer(i->second.lpKey);
	m_advises.erase(i);
	return hrSuccess;
}

HRESULT M4LMAPISupport::Notify(const NOTIFKEY *lpKey, ULONG cNotification,
    NOTIFICATION *lpNotifications, ULONG *lpulFlags)
{
	object_ptr<IMAPIAdviseSink> lpAdviseSink;
	ulock_normal l_adv(m_advises_mutex);

	auto iter = std::find_if(m_advises.cbegin(), m_advises.cend(),
		[=](const auto &entry) {
			return entry.second.lpKey->cb == lpKey->cb &&
			       memcmp(entry.second.lpKey->ab, lpKey->ab, lpKey->cb) == 0;
		});
	if (iter == m_advises.cend())
		/* Should this be reported as error? */
		return hrSuccess;
	lpAdviseSink.reset(iter->second.lpAdviseSink);
	l_adv.unlock();
	return lpAdviseSink->OnNotify(cNotification, lpNotifications);
}

HRESULT M4LMAPISupport::ModifyStatusRow(ULONG cValues,
    const SPropValue *lpColumnVals, ULONG ulFlags)
{
	return static_cast<M4LMAPISession *>(session)->setStatusRow(cValues, lpColumnVals);
}

HRESULT M4LMAPISupport::OpenProfileSection(const MAPIUID *lpUid, ULONG ulFlags,
    IProfSect **lppProfileObj)
{
	if (lpUid == NULL)
		lpUid = lpsProviderUID.get();
	return session->OpenProfileSection(lpUid, nullptr, ulFlags, lppProfileObj);
}

HRESULT M4LMAPISupport::NewUID(LPMAPIUID lpMuid) {
	return CoCreateGuid(reinterpret_cast<GUID *>(lpMuid));
}

HRESULT M4LMAPISupport::CreateOneOff(const TCHAR *lpszName,
    const TCHAR *lpszAdrType, const TCHAR *lpszAddress, ULONG ulFlags,
    ULONG *lpcbEntryID, ENTRYID **lppEntryID)
{
	// although it's called EC... the return value is HRESULT :)
	return ECCreateOneOff(lpszName, lpszAdrType, lpszAddress, ulFlags,
	       lpcbEntryID, lppEntryID);
}

HRESULT M4LMAPISupport::SetProviderUID(const MAPIUID *, ULONG flags)
{
    return hrSuccess;
}

HRESULT M4LMAPISupport::CompareEntryIDs(ULONG cbEntry1, const ENTRYID *lpEntry1,
    ULONG cbEntry2, const ENTRYID *lpEntry2, ULONG ulCompareFlags,
    ULONG *lpulResult)
{
	if (cbEntry1 != cbEntry2)
		*lpulResult = false;
	else if (!lpEntry1 || !lpEntry2)
		return MAPI_E_INVALID_ENTRYID;
	else if (memcmp(lpEntry1, lpEntry2, cbEntry1) != 0)
		*lpulResult = false;
	else
		*lpulResult = true;
	return hrSuccess;
}

HRESULT M4LMAPISupport::OpenEntry(ULONG cbEntryID, const ENTRYID *lpEntryID,
    const IID *lpInterface, ULONG ulOpenFlags, ULONG *lpulObjType,
    IUnknown **lppUnk)
{
	return session->OpenEntry(cbEntryID, lpEntryID, lpInterface,
	       ulOpenFlags, lpulObjType, lppUnk);
}

HRESULT M4LMAPISupport::CopyMessages(const IID *lpSrcInterface,
    void *lpSrcFolder, const ENTRYLIST *lpMsgList, const IID *lpDestInterface,
    void *lpDestFolder, ULONG_PTR ulUIParam, IMAPIProgress *lpProgress,
    ULONG ulFlags)
{
	ULONG ulObjType;
	memory_ptr<ENTRYLIST> lpDeleteEntries;
	bool bPartial = false;

	if (lpSrcInterface == nullptr || lpSrcFolder == nullptr ||
	    lpDestFolder == nullptr || lpMsgList == nullptr)
		return MAPI_E_INVALID_PARAMETER;
	if (*lpSrcInterface != IID_IMAPIFolder ||
	    (lpDestInterface != nullptr && *lpDestInterface != IID_IMAPIFolder))
		return MAPI_E_INTERFACE_NOT_SUPPORTED;

	auto lpSource = static_cast<IMAPIFolder *>(lpSrcFolder);
	auto lpDest = static_cast<IMAPIFolder *>(lpDestFolder);
	auto hr = MAPIAllocateBuffer(sizeof(ENTRYLIST), &~lpDeleteEntries);
	if (hr != hrSuccess)
		return hr;
	hr = MAPIAllocateMore(sizeof(SBinary) * lpMsgList->cValues, lpDeleteEntries,
	     reinterpret_cast<void **>(&lpDeleteEntries->lpbin));
	if (hr != hrSuccess)
		return hr;

	lpDeleteEntries->cValues = 0;
	for (unsigned int i = 0; i < lpMsgList->cValues; ++i) {
		object_ptr<IMessage> lpSrcMessage, lpDestMessage;

		hr = lpSource->OpenEntry(lpMsgList->lpbin[i].cb,
		     reinterpret_cast<ENTRYID *>(lpMsgList->lpbin[i].lpb),
		     &IID_IMessage, 0, &ulObjType, &~lpSrcMessage);
		if (hr != hrSuccess) {
			// partial, or error to calling client?
			bPartial = true;
			continue;
		}
		hr = lpDest->CreateMessage(&IID_IMessage, MAPI_MODIFY, &~lpDestMessage);
		if (hr != hrSuccess) {
			bPartial = true;
			continue;
		}
		hr = DoCopyTo(&IID_IMessage, lpSrcMessage, 0, nullptr, nullptr, ulUIParam, lpProgress, &IID_IMessage, lpDestMessage, 0, nullptr);
		if (FAILED(hr)) {
			return hr;
		} else if (hr != hrSuccess) {
			bPartial = true;
			continue;
		}

		hr = lpDestMessage->SaveChanges(0);
		if (hr != hrSuccess)
			bPartial = true;
		else if (ulFlags & MAPI_MOVE)
			lpDeleteEntries->lpbin[lpDeleteEntries->cValues++] = lpMsgList->lpbin[i];
	}

	if ((ulFlags & MAPI_MOVE) && lpDeleteEntries->cValues > 0 &&
	    lpSource->DeleteMessages(lpDeleteEntries, 0, NULL, 0) != hrSuccess)
		bPartial = true;
	if (bPartial)
		hr = MAPI_W_PARTIAL_COMPLETION;
	return hr;
}

HRESULT M4LMAPISupport::CopyFolder(const IID *lpSrcInterface, void *lpSrcFolder,
    ULONG cbEntryID, const ENTRYID *lpEntryID, const IID *lpDestInterface,
    void *lpDestFolder, const TCHAR *lpszNewFolderName, ULONG_PTR ulUIParam,
    IMAPIProgress *lpProgress, ULONG ulFlags)
{
	object_ptr<IMAPIFolder> lpFolder, lpSubFolder;
	memory_ptr<SPropValue> lpSourceName;
	unsigned int ulObjType = 0, ulFolderFlags = 0;
	static constexpr SizedSPropTagArray (1, sExcludeProps) = {1, {PR_DISPLAY_NAME_A}};

	if (lpSrcInterface == nullptr || lpSrcFolder == nullptr ||
	    cbEntryID == 0 || lpEntryID == nullptr || lpDestFolder == nullptr)
		return MAPI_E_INVALID_PARAMETER;
	if (*lpSrcInterface != IID_IMAPIFolder)
		return MAPI_E_INTERFACE_NOT_SUPPORTED;
	/*
	 * lpDestInterface == NULL or IID_IMAPIFolder compatible.
	 * [Since IMAPIFolder has no known I* descendants, there is
	 * just this one class to handle.]
	 */
	if (lpDestInterface != nullptr && *lpDestInterface != IID_IMAPIFolder)
		return MAPI_E_INTERFACE_NOT_SUPPORTED;
	auto lpSource = static_cast<IMAPIFolder *>(lpSrcFolder);
	auto lpDest = static_cast<IMAPIFolder *>(lpDestFolder);
	auto hr = lpSource->OpenEntry(cbEntryID, lpEntryID, &IID_IMAPIFolder, 0, &ulObjType, &~lpFolder);
	if (hr != hrSuccess)
		return hr;
	if (!lpszNewFolderName) {
		hr = HrGetOneProp(lpFolder, PR_DISPLAY_NAME_W, &~lpSourceName);
		if (hr != hrSuccess)
			return hr;
		ulFolderFlags |= MAPI_UNICODE;
		lpszNewFolderName = (LPTSTR)lpSourceName->Value.lpszW;
	} else
		ulFolderFlags |= (ulFlags & MAPI_UNICODE);

	hr = lpDest->CreateFolder(FOLDER_GENERIC, lpszNewFolderName, nullptr, &IID_IMAPIFolder, ulFolderFlags, &~lpSubFolder);
	if (hr != hrSuccess)
		return hr;
	hr = DoCopyTo(&IID_IMAPIFolder, lpFolder, 0, nullptr, sExcludeProps,
	     ulUIParam, lpProgress, &IID_IMAPIFolder, lpSubFolder, ulFlags, NULL);
	if (hr != hrSuccess)
		return hr;
	if (ulFlags & MAPI_MOVE)
		lpSource->DeleteFolder(cbEntryID, (LPENTRYID)lpEntryID, 0, NULL, DEL_FOLDERS | DEL_MESSAGES);
	return hrSuccess;
}

HRESULT M4LMAPISupport::DoCopyTo(LPCIID lpSrcInterface, LPVOID lpSrcObj,
    ULONG ciidExclude, LPCIID rgiidExclude, const SPropTagArray *lpExcludeProps,
    unsigned int ulUIParam, IMAPIProgress *lpProgress, const IID *lpDestInterface,
    void *lpDestObj, ULONG ulFlags, SPropProblemArray **lppProblems)
{
	return Util::DoCopyTo(lpSrcInterface, lpSrcObj, ciidExclude,
	       rgiidExclude, lpExcludeProps, ulUIParam, lpProgress,
	       lpDestInterface, lpDestObj, ulFlags, lppProblems);
}

HRESULT M4LMAPISupport::DoCopyProps(LPCIID lpSrcInterface, void *lpSrcObj,
    const SPropTagArray *lpIncludeProps, ULONG ulUIParam,
    IMAPIProgress *lpProgress, const IID *lpDestInterface, void *lpDestObj,
    ULONG ulFlags, SPropProblemArray **lppProblems)
{
	return Util::DoCopyProps(lpSrcInterface, lpSrcObj, lpIncludeProps,
	       ulUIParam, lpProgress, lpDestInterface, lpDestObj, ulFlags,
	       lppProblems);
}

/** 
 * Should perform the following tasks:
 *
 * 1 Expand certain personal distribution lists to their component recipients.
 * 2 Replace all display names that have been changed with the original names.
 * 3 Mark any duplicate entries.
 * 4 Resolve all one-off addresses. 
 * 5 Check whether the message needs preprocessing and, if it does,
 *   set the flag pointed to by lpulFlags to NEEDS_PREPROCESSING.
 * 
 * Currently we only do step 1. The rest is done by the spooler and inetmapi.
 *
 * @param[in] lpMessage The message to process
 * @param[out] lpulFlags Return flags for actions required by the caller.
 * 
 * @return MAPI Error code
 */
HRESULT M4LMAPISupport::ExpandRecips(LPMESSAGE lpMessage, ULONG * lpulFlags) {
	object_ptr<IAddrBook> ptrAddrBook;
	std::set<std::string> setFilter;

	auto hr = session->OpenAddressBook(0, nullptr, AB_NO_DIALOG, &~ptrAddrBook);
	if (hr != hrSuccess)
		return hr;
	object_ptr<IMAPITable> ptrRecipientTable;
	hr = lpMessage->GetRecipientTable(fMapiUnicode | MAPI_DEFERRED_ERRORS, &~ptrRecipientTable);
	if (hr != hrSuccess)
		return hr;
	memory_ptr<SPropTagArray> ptrColumns;
	hr = ptrRecipientTable->QueryColumns(TBL_ALL_COLUMNS, &~ptrColumns);
	if (hr != hrSuccess)
		return hr;
	hr = ptrRecipientTable->SetColumns(ptrColumns, 0);
	if (hr != hrSuccess)
		return hr;

	while (true) {
		ULONG ulObjType;
		object_ptr<IDistList> ptrDistList;
		rowset_ptr ptrRow;

		hr = ptrRecipientTable->QueryRows(1, 0L, &~ptrRow);
		if (hr != hrSuccess)
			return hr;
		if (ptrRow.size() == 0)
			break;
		auto lpAddrType = ptrRow[0].cfind(PR_ADDRTYPE);
		if (!lpAddrType)
			continue;
		if (_tcscmp(lpAddrType->Value.LPSZ, KC_T("MAPIPDL")))
			continue;
		auto lpDLEntryID = ptrRow[0].cfind(PR_ENTRYID);
		if (!lpDLEntryID)
			continue;

		if (setFilter.find(std::string(lpDLEntryID->Value.bin.lpb, lpDLEntryID->Value.bin.lpb + lpDLEntryID->Value.bin.cb)) != setFilter.end()) {
			// already expanded this group so continue without opening
			hr = lpMessage->ModifyRecipients(MODRECIP_REMOVE, reinterpret_cast<ADRLIST *>(ptrRow.get()));
			if (hr != hrSuccess)
				return hr;
			continue;
		}
		setFilter.emplace(lpDLEntryID->Value.bin.lpb, lpDLEntryID->Value.bin.lpb + lpDLEntryID->Value.bin.cb);
		hr = ptrAddrBook->OpenEntry(lpDLEntryID->Value.bin.cb,
		     reinterpret_cast<ENTRYID *>(lpDLEntryID->Value.bin.lpb),
		     &iid_of(ptrDistList), 0, &ulObjType, &~ptrDistList);
		if (hr != hrSuccess)
			continue;

		// remove MAPIPDL entry when distlist is opened
		hr = lpMessage->ModifyRecipients(MODRECIP_REMOVE, reinterpret_cast<ADRLIST *>(ptrRow.get()));
		if (hr != hrSuccess)
			return hr;
		object_ptr<IMAPITable> ptrMemberTable;
		hr = ptrDistList->GetContentsTable(fMapiUnicode, &~ptrMemberTable);
		if (hr != hrSuccess)
			continue;

		// Same columns as recipient table
		hr = ptrMemberTable->SetColumns(ptrColumns, 0);
		if (hr != hrSuccess)
			return hr;

		// Get all recipients in distlist, and add to message.
		// If another distlist is here, it will expand in the next loop.
		rowset_ptr ptrMembers;
		hr = ptrMemberTable->QueryRows(INT_MAX, fMapiUnicode, &~ptrMembers);
		if (hr != hrSuccess)
			continue;

		// find all unknown properties in the rows, reference-copy those from the original recipient
		// ModifyRecipients() will actually copy the data
		for (ULONG c = 0; c < ptrMembers.size(); ++c) {
			for (ULONG i = 0; i < ptrMembers[c].cValues; ++i) {
				if (PROP_TYPE(ptrMembers[c].lpProps[i].ulPropTag) != PT_ERROR)
					continue;

				// prop is unknown, find prop in recip, and copy value
				auto lpRecipProp = ptrRow[0].cfind(CHANGE_PROP_TYPE(ptrMembers[c].lpProps[i].ulPropTag, PT_UNSPECIFIED));
				if (lpRecipProp)
					ptrMembers[c].lpProps[i] = *lpRecipProp;
				// else: leave property unknown
			}
		}

		hr = lpMessage->ModifyRecipients(MODRECIP_ADD, reinterpret_cast<ADRLIST *>(ptrMembers.get()));
		if (hr != hrSuccess)
			return hr;
	}
	// Return 0 (no spooler needed, no preprocessing needed)
	if(lpulFlags)
		*lpulFlags = 0;
	return hrSuccess;
}

HRESULT M4LMAPISupport::DoSentMail(ULONG ulFlags, LPMESSAGE lpMessage) {
	return ::DoSentMail(session, nullptr, ulFlags,
	       object_ptr<IMessage>(lpMessage)); /* from CommonUtil */
}

HRESULT M4LMAPISupport::WrapStoreEntryID(ULONG cbOrigEntry,
    const ENTRYID *lpOrigEntry, ULONG *lpcbWrappedEntry,
    ENTRYID **lppWrappedEntry)
{
	if (service == nullptr)
		// addressbook provider doesn't have the SVCService object
		return MAPI_E_CALL_FAILED;
	auto lpDLLName = service->GetProp(PR_SERVICE_DLL_NAME_A);
	if (lpDLLName == nullptr || lpDLLName->Value.lpszA == nullptr)
		return MAPI_E_NOT_FOUND;
	// call mapiutil version
	return ::WrapStoreEntryID(0, reinterpret_cast<TCHAR *>(lpDLLName->Value.lpszA),
	       cbOrigEntry, lpOrigEntry, lpcbWrappedEntry, lppWrappedEntry);
}

HRESULT M4LMAPISupport::ModifyProfile(ULONG ulFlags) {
	return hrSuccess;
}

HRESULT M4LMAPISupport::QueryInterface(const IID &refiid, void **lppInterface)
{
	REGISTER_INTERFACE3(IMAPISup, IMAPISupport, this);
	REGISTER_INTERFACE2(IUnknown, this);
	if (refiid == IID_IMAPIGetSession) {
		IMAPIGetSession *lpGetSession = new M4LMAPIGetSession(session);
		lpGetSession->AddRef();
		*lppInterface = lpGetSession;
		return hrSuccess;
	}
	return MAPI_E_INTERFACE_NOT_SUPPORTED;
}
