/*
 * SPDX-License-Identifier: AGPL-3.0-only
 * Copyright 2005 - 2016 Zarafa and its licensors
 */
#include <stdexcept>
#include <vector>
#include <kopano/platform.h>
#include <kopano/memory.hpp>
#include <kopano/scope.hpp>
#include "WSMAPIFolderOps.h"
#include <kopano/ECGuid.h>
#include "SOAPSock.h"
#include "SOAPUtils.h"
#include "WSUtil.h"
#include <kopano/charset/utf8string.h>

#define START_SOAP_CALL retry: \
	if (m_lpTransport->m_lpCmd == nullptr) \
		return MAPI_E_NETWORK_ERROR;
#define END_SOAP_CALL 	\
	if (er == KCERR_END_OF_SESSION && m_lpTransport->HrReLogon() == hrSuccess) \
		goto retry; \
	hr = kcerr_to_mapierr(er, MAPI_E_NOT_FOUND); \
	if(hr != hrSuccess) \
		return hr;

/*
 * The WSMAPIFolderOps for use with the WebServices transport
 */
using namespace KC;

WSMAPIFolderOps::WSMAPIFolderOps(ECSESSIONID sid, ULONG cbEntryId,
    const ENTRYID *lpEntryId, WSTransport *lpTransport) :
	ecSessionId(sid), m_lpTransport(lpTransport)
{
	lpTransport->AddSessionReloadCallback(this, Reload, &m_ulSessionReloadCallback);
	auto ret = CopyMAPIEntryIdToSOAPEntryId(cbEntryId, lpEntryId, &m_sEntryId);
	if (ret != hrSuccess)
		throw std::runtime_error("CopyMAPIEntryIdToSOAPEntryId");
}

WSMAPIFolderOps::~WSMAPIFolderOps()
{
	m_lpTransport->RemoveSessionReloadCallback(m_ulSessionReloadCallback);
	soap_del_entryId(&m_sEntryId);
}

HRESULT WSMAPIFolderOps::Create(ECSESSIONID ecSessionId, ULONG cbEntryId,
    const ENTRYID *lpEntryId, WSTransport *lpTransport,
    WSMAPIFolderOps **lppFolderOps)
{
	return alloc_wrap<WSMAPIFolderOps>(ecSessionId, cbEntryId, lpEntryId,
	       lpTransport).put(lppFolderOps);
}

HRESULT WSMAPIFolderOps::QueryInterface(REFIID refiid, void **lppInterface)
{
	REGISTER_INTERFACE3(ECMAPIFolderOps, WSMAPIFolderOps, this);
	REGISTER_INTERFACE2(ECUnknown, this);
	REGISTER_INTERFACE2(IUnknown, this);
	return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

HRESULT WSMAPIFolderOps::HrCreateFolder(ULONG ulFolderType,
    const utf8string &strFolderName, const utf8string &strComment,
    BOOL fOpenIfExists, ULONG ulSyncId, const SBinary *lpsSourceKey,
    ULONG cbNewEntryId, LPENTRYID lpNewEntryId, ULONG* lpcbEntryId,
    LPENTRYID *lppEntryId)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;
	struct xsd__base64Binary sSourceKey;
	entryId*	lpsEntryId = NULL;
	auto cleanup = make_scope_exit([&]() { soap_del_PointerToentryId(&lpsEntryId); });

	if(lpNewEntryId) {
		hr = CopyMAPIEntryIdToSOAPEntryId(cbNewEntryId, lpNewEntryId, &lpsEntryId);
		if(hr != hrSuccess)
			return hr;
	}
	if (lpsSourceKey) {
		sSourceKey.__ptr = lpsSourceKey->lpb;
		sSourceKey.__size = lpsSourceKey->cb;
	} else {
		sSourceKey.__ptr = NULL;
		sSourceKey.__size = 0;
	}

	soap_lock_guard spg(*m_lpTransport);
	struct createFolderResponse sResponse;
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->createFolder(ecSessionId,
		    m_sEntryId, lpsEntryId, ulFolderType, strFolderName.z_str(),
		    strComment.z_str(), !!fOpenIfExists, ulSyncId, sSourceKey,
		    &sResponse) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	if(lpcbEntryId != NULL && lppEntryId != NULL)
		hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sEntryId, lpcbEntryId, lppEntryId);
	return hr;
}

HRESULT WSMAPIFolderOps::create_folders(std::vector<WSFolder> &batch)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	struct create_folders_response rsp;
	std::vector<new_folder> folders(batch.size());
	auto cleanup = make_scope_exit([&]() {
		for (auto &folder : folders)
			if (folder.entryid != nullptr)
				soap_del_PointerToentryId(&folder.entryid);
	});
	convert_wsfolder_to_soapfolder(batch, folders);

	new_folder_set soap_batch;
	soap_batch.__size = folders.size();
	soap_batch.__ptr  = folders.data();

	soap_lock_guard spg(*m_lpTransport);
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->create_folders(ecSessionId, m_sEntryId, soap_batch, &rsp) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
		else
			er = rsp.er;
	}
	END_SOAP_CALL
	if (rsp.entryids == nullptr || rsp.entryids->__size != batch.size())
		return MAPI_E_CALL_FAILED;
	return convert_soapfolders_to_wsfolder(rsp, batch);
}

HRESULT WSMAPIFolderOps::HrDeleteFolder(ULONG cbEntryId,
    const ENTRYID *lpEntryId, ULONG ulFlags, ULONG ulSyncId)
{
	ECRESULT	er = erSuccess;
	entryId		sEntryId;

	// Cheap copy entryid
	auto hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryId, lpEntryId, &sEntryId, true);
	if(hr != hrSuccess)
		return hr;

	soap_lock_guard spg(*m_lpTransport);
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->deleteFolder(ecSessionId, sEntryId,
		    ulFlags, ulSyncId, &er) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
	}
	END_SOAP_CALL
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrEmptyFolder(ULONG ulFlags, ULONG ulSyncId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	soap_lock_guard spg(*m_lpTransport);

	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->emptyFolder(ecSessionId,
		    m_sEntryId, ulFlags, ulSyncId, &er) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
	}
	END_SOAP_CALL
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrSetReadFlags(ENTRYLIST *lpMsgList, ULONG ulFlags, ULONG ulSyncId)
{
	HRESULT			hr = hrSuccess;
	ECRESULT		er = erSuccess;
	struct entryList sEntryList;
	auto cleanup = make_scope_exit([&]() { soap_del_entryList(&sEntryList); });

	if(lpMsgList) {
		if(lpMsgList->cValues == 0)
			return hr;
		hr = CopyMAPIEntryListToSOAPEntryList(lpMsgList, &sEntryList);
		if(hr != hrSuccess)
			return hr;
	}

	soap_lock_guard spg(*m_lpTransport);
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->setReadFlags(ecSessionId, ulFlags,
		    &m_sEntryId, lpMsgList != nullptr ? &sEntryList : nullptr,
		    ulSyncId, &er) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
	}
	END_SOAP_CALL
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrSetSearchCriteria(const ENTRYLIST *lpMsgList,
    const SRestriction *lpRestriction, ULONG ulFlags)
{
	ECRESULT		er = erSuccess;
	HRESULT			hr = hrSuccess;
	struct entryList*		lpsEntryList = NULL;
	struct restrictTable*	lpsRestrict = NULL;
	auto cleanup = make_scope_exit([&] () {
		soap_del_PointerTorestrictTable(&lpsRestrict);
		soap_del_PointerToentryList(&lpsEntryList);
	});

	if(lpMsgList) {
		lpsEntryList = soap_new_entryList(nullptr);
		hr = CopyMAPIEntryListToSOAPEntryList(lpMsgList, lpsEntryList);
		if(hr != hrSuccess)
			return hr;
	}
	if (lpRestriction)
	{
		hr = CopyMAPIRestrictionToSOAPRestriction(&lpsRestrict, lpRestriction);
		if(hr != hrSuccess)
			return hr;
	}

	soap_lock_guard spg(*m_lpTransport);
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->tableSetSearchCriteria(ecSessionId,
		    m_sEntryId, lpsRestrict, lpsEntryList, ulFlags, &er) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
	}
	END_SOAP_CALL
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrGetSearchCriteria(ENTRYLIST **lppMsgList, LPSRestriction *lppRestriction, ULONG *lpulFlags)
{
	HRESULT			hr = hrSuccess;
	ECRESULT		er = erSuccess;
	memory_ptr<ENTRYLIST> lpMsgList;
	memory_ptr<SRestriction> lpRestriction;
	soap_lock_guard spg(*m_lpTransport);
	struct tableGetSearchCriteriaResponse sResponse;

	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->tableGetSearchCriteria(ecSessionId,
		    m_sEntryId, &sResponse) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	if(lppRestriction) {
		hr = MAPIAllocateBuffer(sizeof(SRestriction), &~lpRestriction);
		if(hr != hrSuccess)
			return hr;
		hr = CopySOAPRestrictionToMAPIRestriction(lpRestriction, sResponse.lpRestrict, lpRestriction);
		if(hr != hrSuccess)
			return hr;
	}
	if(lppMsgList) {
		hr = CopySOAPEntryListToMAPIEntryList(sResponse.lpFolderIDs, &~lpMsgList);
		if(hr != hrSuccess)
			return hr;
	}
	if(lppMsgList)
		*lppMsgList = lpMsgList.release();
	if(lppRestriction)
		*lppRestriction = lpRestriction.release();
	if(lpulFlags)
		*lpulFlags = sResponse.ulFlags;
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrCopyFolder(ULONG cbEntryFrom,
    const ENTRYID *lpEntryFrom, ULONG cbEntryDest, const ENTRYID *lpEntryDest,
    const utf8string &strNewFolderName, ULONG ulFlags, ULONG ulSyncId)
{
	ECRESULT	er = erSuccess;
	entryId sEntryFrom, sEntryDest; /* Do not free, cheap copy */

	auto hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryFrom, lpEntryFrom, &sEntryFrom, true);
	if(hr != hrSuccess)
		return hr;
	hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryDest, lpEntryDest, &sEntryDest, true);
	if(hr != hrSuccess)
		return hr;

	soap_lock_guard spg(*m_lpTransport);
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->copyFolder(ecSessionId, sEntryFrom, sEntryDest,
		    strNewFolderName.z_str(), ulFlags, ulSyncId, &er) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
	}
	END_SOAP_CALL
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrCopyMessage(ENTRYLIST *lpMsgList, ULONG cbEntryDest,
    const ENTRYID *lpEntryDest, ULONG ulFlags, ULONG ulSyncId)
{
	if (lpMsgList->cValues == 0)
		return hrSuccess;

	ECRESULT		er = erSuccess;
	struct entryList sEntryList;
	auto cleanup = make_scope_exit([&]() { soap_del_entryList(&sEntryList); });
	entryId			sEntryDest;	//Do not free, cheap copy

	auto hr = CopyMAPIEntryListToSOAPEntryList(lpMsgList, &sEntryList);
	if(hr != hrSuccess)
		return hr;
	hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryDest, lpEntryDest, &sEntryDest, true);
	if(hr != hrSuccess)
		return hr;

	soap_lock_guard spg(*m_lpTransport);
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->copyObjects(ecSessionId,
		    &sEntryList, sEntryDest, ulFlags, ulSyncId, &er) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
	}
	END_SOAP_CALL
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrGetMessageStatus(ULONG cbEntryID,
    const ENTRYID *lpEntryID, ULONG ulFlags, ULONG *lpulMessageStatus)
{
	if (lpEntryID == nullptr)
		return MAPI_E_INVALID_ENTRYID;

	ECRESULT	er = erSuccess;
	entryId		sEntryId;	//Do not free, cheap copy
	auto hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryID, lpEntryID, &sEntryId, true);
	if(hr != hrSuccess)
		return hr;

	soap_lock_guard spg(*m_lpTransport);
	struct messageStatus sMessageStatus;
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->getMessageStatus(ecSessionId,
		    sEntryId, ulFlags, &sMessageStatus) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
		else
			er = sMessageStatus.er;
	}
	END_SOAP_CALL
	*lpulMessageStatus = sMessageStatus.ulMessageStatus;
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrSetMessageStatus(ULONG cbEntryID,
    const ENTRYID *lpEntryID, ULONG ulNewStatus, ULONG ulNewStatusMask,
    ULONG ulSyncId, ULONG *lpulOldStatus)
{
	if (lpEntryID == nullptr)
		return MAPI_E_INVALID_ENTRYID;

	ECRESULT	er = erSuccess;
	entryId		sEntryId;	//Do not free, cheap copy
	auto hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryID, lpEntryID, &sEntryId, true);
	if(hr != hrSuccess)
		return hr;

	soap_lock_guard spg(*m_lpTransport);
	struct messageStatus sMessageStatus;
	START_SOAP_CALL
	{
		if (m_lpTransport->m_lpCmd->setMessageStatus(ecSessionId,
		    sEntryId, ulNewStatus, ulNewStatusMask, ulSyncId,
		    &sMessageStatus) != SOAP_OK)
			er = KCERR_NETWORK_ERROR;
		else
			er = sMessageStatus.er;
	}
	END_SOAP_CALL

	if(lpulOldStatus)
		*lpulOldStatus = sMessageStatus.ulMessageStatus;
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::HrGetChangeInfo(ULONG cbEntryID,
    const ENTRYID *lpEntryID, SPropValue **lppPropPCL, SPropValue **lppPropCK)
{
	if (lpEntryID == nullptr)
		return MAPI_E_INVALID_ENTRYID;

	ECRESULT	er = erSuccess;
	entryId sEntryId;
	memory_ptr<SPropValue> lpSPropValPCL, lpSPropValCK;
	auto hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryID, lpEntryID, &sEntryId, true);
	if(hr != hrSuccess)
		return hr;

	soap_lock_guard spg(*m_lpTransport);
	getChangeInfoResponse sChangeInfo;
	if (m_lpTransport->m_lpCmd == nullptr ||
	    m_lpTransport->m_lpCmd->getChangeInfo(ecSessionId, sEntryId, &sChangeInfo) != SOAP_OK)
		er = KCERR_NETWORK_ERROR;
	else
		er = sChangeInfo.er;

	hr = kcerr_to_mapierr(er);
	if (hr != hrSuccess)
		return hr;

	if (lppPropPCL) {
		hr = MAPIAllocateBuffer(sizeof *lpSPropValPCL, &~lpSPropValPCL);
		if (hr != hrSuccess)
			return hr;
		hr = CopySOAPPropValToMAPIPropVal(lpSPropValPCL, &sChangeInfo.sPropPCL, lpSPropValPCL);
		if (hr != hrSuccess)
			return hr;
	}

	if (lppPropCK) {
		hr = MAPIAllocateBuffer(sizeof *lpSPropValCK, &~lpSPropValCK);
		if (hr != hrSuccess)
			return hr;
		hr = CopySOAPPropValToMAPIPropVal(lpSPropValCK, &sChangeInfo.sPropCK, lpSPropValCK);
		if (hr != hrSuccess)
			return hr;
	}

	// All went well, modify output parameters
	if (lppPropPCL != nullptr)
		*lppPropPCL = lpSPropValPCL.release();
	if (lppPropCK != nullptr)
		*lppPropCK = lpSPropValCK.release();
	return hrSuccess;
}

HRESULT WSMAPIFolderOps::Reload(void *lpParam, ECSESSIONID sessionid)
{
	static_cast<WSMAPIFolderOps *>(lpParam)->ecSessionId = sessionid;
	return hrSuccess;
}
