/*
 * SPDX-License-Identifier: AGPL-3.0-only
 * Copyright 2005 - 2016 Zarafa and its licensors
 */
#include <kopano/platform.h>
#include <memory>
#include <cerrno>
#include <cstring>
#include <sys/un.h>
#include <libHX/string.h>
#include "SOAPSock.h"
#include "SOAPUtils.h"
#include <kopano/ECLogger.h>
#include <kopano/stringutil.h>
#include <kopano/CommonUtil.h>
#include <kopano/charset/convert.h>

using namespace KC;

KCmdProxy2::~KCmdProxy2()
{
	free(const_cast<char *>(soap_endpoint));
#if GSOAP_VERSION < 20875
	destroy();
#endif
}

// we cannot patch http_post now (see external/gsoap/*.diff), so we redefine it
static int
http_post(struct soap *soap, const char *endpoint, const char *host, int port, const char *path, const char *action, ULONG64 count)
{ int err;
  if (strlen(endpoint) + strlen(soap->http_version) > sizeof(soap->tmpbuf) - 80
   || strlen(host) + strlen(soap->http_version) > sizeof(soap->tmpbuf) - 80)
    return soap->error = SOAP_EOM;
  sprintf(soap->tmpbuf, "POST /%s HTTP/%s", (*path == '/' ? path + 1 : path), soap->http_version);

	err = soap->fposthdr(soap, soap->tmpbuf, nullptr);
	if (err != 0)
		return err;
	err = soap->fposthdr(soap, "Host", host);
	if (err != 0)
		return err;
	err = soap->fposthdr(soap, "User-Agent", "gSOAP/2.8");
	if (err != 0)
		return err;
	err = soap_puthttphdr(soap, SOAP_OK, count);
	if (err != 0)
		return err;
#ifdef WITH_ZLIB
#ifdef WITH_GZIP
	err = soap->fposthdr(soap, "Accept-Encoding", "gzip, deflate");
#else
	err = soap->fposthdr(soap, "Accept-Encoding", "deflate");
#endif
	if (err != 0)
		return err;
#endif
  return soap->fposthdr(soap, NULL, NULL);
}

// This function wraps the GSOAP fopen call to support "file:///var/run/socket" Unix socket URIs
static int gsoap_connect_pipe(struct soap *soap, const char *endpoint,
    const char *host, int port)
{
	// See stdsoap2.cpp:tcp_connect() function
	if (soap_valid_socket(soap->socket))
	    return SOAP_OK;

	struct sockaddr_un saddr;
	memset(&saddr, 0, sizeof(struct sockaddr_un));
	soap->socket = SOAP_INVALID_SOCKET;
	soap->errmode = 0;

	if (strncmp(endpoint, "file:", 5) != 0) {
		soap_set_sender_error(soap, strerror(EINVAL), "file: scheme is required for gsoap_connect_pipe", SOAP_TCP_ERROR);
		return SOAP_INVALID_SOCKET;
	}
	const char *socket_name = strchr(endpoint + 7, '/');
	// >= because there also needs to be room for the 0x00
	if (socket_name == NULL ||
	    strlen(socket_name) >= sizeof(saddr.sun_path)) {
		soap_set_sender_error(soap, strerror(EINVAL), "Socket path is too long", SOAP_TCP_ERROR);
		return SOAP_INVALID_SOCKET;
	}
	auto fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd < 0) {
#if GSOAP_VERSION >= 208103
		soap->errnum = soap_socket_errno;
#else
		soap->errnum = soap_socket_errno(fd);
#endif
		soap_set_sender_error(soap, strerror(errno), "connect_pipe: socket failed", SOAP_TCP_ERROR);
		return SOAP_INVALID_SOCKET;
	}

	saddr.sun_family = AF_UNIX;
	HX_strlcpy(saddr.sun_path, socket_name, sizeof(saddr.sun_path));
	if (connect(fd, (struct sockaddr *)&saddr, sizeof(struct sockaddr_un)) < 0) {
#if GSOAP_VERSION >= 208103
		soap->errnum = soap_socket_errno;
#else
		soap->errnum = soap_socket_errno(fd);
#endif
		soap_set_receiver_error(soap, strerror(errno), "connect_pipe: connect failed", SOAP_TCP_ERROR);
		close(fd);
		return SOAP_INVALID_SOCKET;
	}

 	soap->sendfd = soap->recvfd = SOAP_INVALID_SOCKET;
	soap->socket = fd;
	// Because 'file:///var/run/file' will be parsed into host='', path='/var/run/file',
	// the gSOAP code doesn't set the soap->status variable. (see soap_connect_command in
	// stdsoap2.cpp:12278) This could possibly lead to
	// soap->status accidentally being set to SOAP_GET, which would break things. The
	// chances of this happening are, of course, small, but also very real.
	soap->status = SOAP_POST;
	/* Do like gsoap's tcp_connect would */
	soap->keep_alive = -((soap->omode & SOAP_IO_KEEPALIVE) != 0);
   	return SOAP_OK;
}

HRESULT CreateSoapTransport(const sGlobalProfileProps &prof, KCmdProxy2 **lppCmd)
{
	if (prof.strServerPath.size() == 0 || lppCmd == nullptr)
		return E_INVALIDARG;
	auto lpCmd = make_unique_nt<KCmdProxy2>(SOAP_IO_KEEPALIVE | SOAP_C_UTFSTRING,
	             SOAP_IO_KEEPALIVE | SOAP_XML_TREE | SOAP_C_UTFSTRING);
	lpCmd->soap_endpoint = strdup(prof.strServerPath.c_str());
	lpCmd->soap->sndbuf = lpCmd->soap->rcvbuf = 0;
	lpCmd->soap->maxoccurs = SIZE_MAX; // override default limit of 100000, as this breaks ICS for large folders at least
	// default allow SSLv3, TLSv1, TLSv1.1 and TLSv1.2
	lpCmd->soap->ctx = SSL_CTX_new(SSLv23_method());

#ifdef WITH_OPENSSL
	if (strncmp("https:", lpCmd->soap_endpoint, 6) == 0) {
		// no need to add certificates to call, since soap also calls SSL_CTX_set_default_verify_paths()
		if (soap_ssl_client_context(lpCmd->soap, SOAP_SSL_DEFAULT,
		    prof.strSSLKeyFile.size() > 0 ? prof.strSSLKeyFile.c_str() : nullptr,
		    prof.strSSLKeyPass.size() > 0 ? prof.strSSLKeyPass.c_str() : nullptr,
		    nullptr, nullptr, nullptr))
			return E_INVALIDARG;
		// set our own certificate check function
		lpCmd->soap->fsslverify = ssl_verify_callback_kopano_silent;
		SSL_CTX_set_verify(lpCmd->soap->ctx, SSL_VERIFY_PEER, lpCmd->soap->fsslverify);
	}
#endif
	if (*lpCmd->soap_endpoint == '/')
		/*
		 * Based upon "a.com/b" being a common shorthand for
		 * "http://a.com/b", gsoap treats "/b" as a shorthand for
		 * "http://<zero-length-domain>/b". The zero-length domain
		 * however is a (documented) way to emit HTTP requests on
		 * (by default) stdout.
		 */
		return E_INVALIDARG;

	if(strncmp("file:", lpCmd->soap_endpoint, 5) == 0) {
		lpCmd->soap->fconnect = gsoap_connect_pipe;
		lpCmd->soap->fpost = http_post;
	} else {
		if ((prof.ulProxyFlags & 0x0000001 /* EC_PROFILE_PROXY_FLAGS_USE_PROXY */) &&
		    prof.strProxyHost.size() > 0) {
			lpCmd->soap->proxy_host = soap_strdup(lpCmd->soap, prof.strProxyHost.c_str());
			lpCmd->soap->proxy_port = prof.ulProxyPort;
			if (prof.strProxyUserName.size() > 0)
				lpCmd->soap->proxy_userid = soap_strdup(lpCmd->soap, prof.strProxyUserName.c_str());
			if (prof.strProxyPassword.size() > 0)
				lpCmd->soap->proxy_passwd = soap_strdup(lpCmd->soap, prof.strProxyPassword.c_str());
		}
		lpCmd->soap->connect_timeout = prof.ulConnectionTimeOut;
	}
	*lppCmd = lpCmd.release();
	return hrSuccess;
}

int ssl_verify_callback_kopano_silent(int ok, X509_STORE_CTX *store)
{
	if (ok != 0)
		return ok;
	auto sslerr = X509_STORE_CTX_get_error(store);
	switch (sslerr) {
	case X509_V_ERR_CERT_HAS_EXPIRED:
	case X509_V_ERR_CERT_NOT_YET_VALID:
	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
		// always ignore these errors
		X509_STORE_CTX_set_error(store, X509_V_OK);
		ok = 1;
		break;
	default:
		break;
	}
	return ok;
}
